from __future__ import print_function
import functools

_modes = ["push", "pull"]

_types = set(("event", "exception", "int", "float", "bool", "str", "mstr", "id","matrix","object"))

_typemap = {
  "String": "str",
  "Material":"str",
  "Integer": "int",
  "Float": "float",
  "Bool": "bool",
  "FunctionLink": ("str", "function"),
  "ObjectLink": ("str", "object"),
  "ModuleLink": ("str", "module"),
  "Filename": ("str", "filename"),
}


from .event import event
from .mstr import mstr
from .reference import reference

eventparser_todo_warn = True

import spyder

def stringtupleparser(*v):
  if len(v) == 1: v = v[0]
  value = str(v).strip()
  if not len(value): return ""
  if value[0] == "(" and value[-1] == ")":
    value = value[1:-1]
  parenth = 0
  quote = None
  tokens = []
  last = 0
  found = False
  skip = 0
  for lnr, l in enumerate(value):
    if l == "," and quote is None:
      if parenth == 0:
        tokens.append(value[last:lnr-skip])
        last = lnr+1
        continue
        
    skip = 0
    if l == "(":
      parenth += 1
      found = True
      if lnr == last: last += 1
    elif l == ")":
      parenth -= 1
    elif (l == "'" or l == '"'):
     if quote is None and last == lnr:
       last += 1
       quote = l
     elif quote == l:
       quote = None
       skip = 1
    elif l == " " and quote is None and last == lnr:
      last += 1   
  if parenth > 0: return value
  if last < len(value)-skip:
    tokens.append(value[last:len(value)-skip])

  if not found and len(tokens) == 1:
    return tokens[0]
  ret = []
  for tok in tokens: 
    ret.append(stringtupleparser(tok))
  return tuple(ret)

def eventparser(e):
  if isinstance(e, str): 
    ee = e
  elif isinstance(e, tuple):
    ee = stringtupleparser(*e)
  return event(e)
  
def boolparser(v):
  if v in (False, "False", 0, "0"): return False
  if v in (True, "True", 1, "1"): return True
  raise ValueError(v)

def generic_constructor(t=None):
  if t is None:
    def generic_constructor1(v):
      if isinstance(v, reference): return v.obj    
      try:
        return type(v)(v)
      except TypeError:
        return v
    return generic_constructor1
  else:
    def generic_constructor2(v):
      if isinstance(v, reference): return v.obj    
      try:
        return t(v)
      except TypeError:
        return v
    return generic_constructor2
    

_parametertypes = {
  "int": ("int", int),
  "float": ("float", float),
  "bool": ("bool", boolparser),
  "str": ("str", str),
  "mstr":("mstr",generic_constructor(mstr)),
  "id": ("str", str),  
  "object":("str", generic_constructor()), 
  "event":("event",eventparser),
  "matrix":("matrix",generic_constructor()),
}

def parse_parameters(unnamedparameters, namedparameters, args, kargs, exactmatch=True):
  kargs = dict(kargs)
  for p in namedparameters[len(args):]:
    if p[0] not in kargs and p[2] != "no-defaultvalue": kargs[p[0]]=p[2]
  
  wrongargnr = False
  if exactmatch:
    if len(args) + len(kargs) != len(unnamedparameters) + len(namedparameters):
      wrongargnr = True
  else:
    if len(args) + len(kargs) < len(unnamedparameters) + len(namedparameters):
      wrongargnr = True
  if wrongargnr:    
    raise TypeError("""
Number of required parameters: %d\n
Number of supplied parameters: %d (%d non-keyword, %d keyword)
""" % (len(unnamedparameters) + len(namedparameters), len(args)+len(kargs), len(args), len(kargs)))
  if len(args) < len(unnamedparameters):
    raise TypeError("""
Number of required non-keyword parameters: %d\n
Number of supplied non-keyword parameters: %d (%d non-keyword, %d keyword)
""" % (len(unnamedparameters) , len(args)))
  ret1 = []
  ret2 = {}
  for anr, a in enumerate(args):
    if anr == len(unnamedparameters): break
    pclass = unnamedparameters[anr][1]  
    if pclass is object: pval = a
    elif a is None: pval = None
    else: pval = pclass(a)
    ret1.append(pval)
  args = args[len(unnamedparameters):]
  
  params = {}
  for p in enumerate(namedparameters): params[p[0]] = p[1]
  for anr, a in enumerate(args):
    if anr >= len(namedparameters): 
      break
    pname, pclass,default = namedparameters[anr]
    pclass = pclass[1]
    if pclass is object: pval = a
    elif a is None: pval = None
    else: pval = pclass(a)
    ret2[pname] = pval
  
  unmatched = {}
  for pname in kargs:
    if pname in ret2:
      raise TypeError("Duplicate definition of parameter %s" % pname)
    pclass = [v[1] for v in namedparameters if v[0] == pname]
    if len(pclass) == 0:
      unmatched[pname] = kargs[pname]
      continue
    pclass = pclass[0][1]
    a = kargs[pname]
    if pclass is object: pval = a
    elif a is None: pval = None
    else: pval = pclass(a)
    ret2[pname] = pval
  for pname,pclass,default in namedparameters:
    if pname not in ret2:
      raise TypeError("Undefined parameter %s" % pname)
  if exactmatch:
    return tuple(ret1), ret2
  else:
    return tuple(ret1), ret2, unmatched
        
def _parameterclass(parametertuple, *args, **kargs):
  return parse_parameters(parametertuple,[],args[0], kargs)[0]

def get_parameterclass(parclassname):
  t = typeclass(parclassname)
  if t.is_subtype: return get_parameterclass(parclassname[0])
  if isinstance(parclassname, tuple):
    parametertuple = [get_parameterclass(p) for p in parclassname]
    t0 = [t[0] for t in parametertuple]
    t1 = [t[1] for t in parametertuple]
    #return [t0, functools.partial(_parameterclass, t1), "no-defaultvalue"]
    #return [t0, functools.partial(_parameterclass, parametertuple), "no-defaultvalue"]
    return (tuple(t0), functools.partial(_parameterclass, parametertuple), "no-defaultvalue")
  elif parclassname in _typemap: 
    return get_parameterclass(_typemap[parclassname])
  elif spyder.validvar2(parclassname):
    import Spyder
    return (parclassname, getattr(Spyder, parclassname))
  else:
    return tuple(_parametertypes[parclassname]) + ("no-defaultvalue",)

def add_parameterclass(parclassname, parclassexpr, parclass):
  global _parametertypes
  assert not isinstance(parclassname, tuple)
  _parametertypes[parclassname] = (parclassexpr, parclass)

_pushtypes = set(("trigger", "toggle"))

_triggermodes = ["default", "input", "output", "update"]

class types_baseclass(object):
  def __ne__(self, other): return not self.__eq__(other)      
  def __eq__(self, other):
    v = self.value
    ov = other
    if isinstance(other, type(self)): ov = other.value
    if v in _typemap: v = _typemap[v]
    if ov in _typemap: ov = _typemap[ov]
    return v == ov


class modeclass(types_baseclass):
  def __init__(self, arg):
    if isinstance(arg, modeclass):
      self.value = arg.value
    else:
      if arg not in _modes:
        raise TypeError("Unknown channel mode %s" % arg)
      self.value = arg

class typeclass(types_baseclass):
  def __init__(self, arg):
     self.push_type = False
     self.is_subtype = False
     if isinstance(arg, typeclass):
       self.value = arg.value
       self.is_subtype = arg.is_subtype
       self.is_compound = arg.is_compound
       self.push_type = arg.push_type
     elif isinstance(arg, tuple):
       try:
         for t in arg:
           typeclass(t)
         self.is_compound = True       
         self.value = arg
       except TypeError:
         typeclass(arg[0])
         self.is_compound = False
         self.is_subtype = True
         self.value = arg
     else:
       self.check_type(arg)
       self.is_compound = False              
       self.value = arg  
  def check_type(self, t):
    if spyder.validvar2(t): return
    if t not in _types:
      if t in _pushtypes:
        self.push_type = True        
      else:
        raise TypeError("Unknown channel data type %s" % t)
  def get_tuple(self):
    if self.is_compound: return self.value
    else: return (self.value,)

class mode_type(object):
  def __init__(self, mode, type=None):
    if isinstance(mode, mode_type):
      if type != None: raise TypeError("Too many arguments")    
      self.mode = mode.mode
      self.type = mode.type  
    else:
      if type == None: raise TypeError("Too few arguments")    
      self.mode = modeclass(mode)
      self.type = typeclass(type)
    if self.mode.value == "pull" and self.type.push_type:
      raise TypeError("Invalid channel data type %s for pull data" % self.type.value)
  def __ne__(self, other): return not self.__eq__(other)      
  def __eq__(self, other):
    other = mode_type(other)
    if self.mode != other.mode: return False
    if self.type != other.type: return False
    return True

class connection_inputclass(object):
  def __init__(self, arg):
    if isinstance(arg, str):
      self.value = arg
      self.bound = False
    else:
      self.test(arg)
  def test(self, arg):
    if hasattr(arg, "connection_input_type") and hasattr(arg, "connection_input"):
      self.value = arg
      self.bound = True  
    else:
      raise TypeError("Segment %s cannot be used as output from another segment" % type(arg).__name__)        
  def bind(self, classname, dic):
    if not self.bound:
      if self.value not in dic:
        raise KeyError("Segment name %s was not declared in worker class %s" % (self.value, classname))
      self.test(dic[self.value])

class connection_outputclass(object):
  def __init__(self, arg):
    if isinstance(arg, str):
      self.value = arg
      self.bound = False
    else:
      self.test(arg)
  def test(self, arg):
    if hasattr(arg, "connection_output_type") and hasattr(arg, "connection_output"):
      self.value = arg
      self.bound = True  
    else:
      raise TypeError("Segment %s cannot be used as input to another segment" % type(arg).__name__)        
  def bind(self, classname, dic):
    if not self.bound:
      if self.value not in dic:
        raise KeyError("Segment name %s was not declared in worker class %s" % (self.value, classname))
      self.test(dic[self.value])

class triggering_class(object):
  def __init__(self, arg, mode, pre):
    if mode not in _triggermodes:
      raise TypeError("Unknown trigger mode %s" % mode)  
    self.mode = mode
    self.pre = pre
    if isinstance(arg, str):
      self.value = arg
      self.bound = False
    else:
      self.test(arg)
  def test(self, arg):
    if hasattr(arg, "triggering_" + self.mode):
      self.value = arg
      self.bound = True  
    else:
      raise TypeError("Segment %s cannot send triggers in mode %s" % (type(arg).__name__, self.mode))        

    #hack: change default pretrigger mode of variables from "input" to "output"
    from .segments.variable import variable    
    if self.pre and isinstance(self.value, variable) and self.mode == "default":
      self.mode = "output"
    
  def bind(self, classname, dic):
    if not self.bound:
      if self.value not in dic:
        raise KeyError("Segment name %s was not declared in worker class %s" % (self.value, classname))
      self.test(dic[self.value])

class triggered_class(object):
  def __init__(self, arg, mode):
    if mode not in _triggermodes:
      raise TypeError("Unknown trigger mode %s" % mode)  
    self.mode = mode
    if isinstance(arg, str):
      self.value = arg
      self.bound = False
    else:
      self.test(arg)
  def test(self, arg):
    if hasattr(arg, "triggered_" + self.mode):
      self.value = arg
      self.bound = True  
    else:
      raise TypeError("Segment %s cannot receive triggers in mode %s" % (type(arg).__name__, self.mode))        
  def bind(self, classname, dic):
    if not self.bound:
      if self.value not in dic:
        raise KeyError("Segment name %s was not declared in worker class %s" % (self.value, classname))
      self.test(dic[self.value])

class startvalueclass(object):
  def __init__(self, arg):
    if isinstance(arg, str):
      self.value = arg
      self.bound = False
    else:
      self.test(arg)
  def test(self, arg):
    if hasattr(arg, "set_startvalue"):
      self.value = arg
      self.bound = True  
    else:
      raise TypeError("Segment %s cannot accept a start value" % type(arg).__name__)        
  def bind(self, classname, dic):
    if not self.bound:
      if self.value not in dic:
        raise KeyError("Segment name %s was not declared in worker class %s" % (self.value, classname))
      self.test(dic[self.value])

def typecompare(t1, t2):
  try:
    tt1 = typeclass(t1)
    tt2 = typeclass(t2)
  except TypeError:
    return t1 == t2
  if not tt1.is_subtype and not tt2.is_subtype:
    return tt1 == tt2
  else:
    if not tt1.is_subtype: t1 = (t1,)
    if not tt2.is_subtype: t2 = (t2,)
    for v1,v2 in zip(t1,t2):
      if v1 != v2: return False
    return True
