from __future__ import print_function, absolute_import

"""
Builds worker templates and instantiates them with get_copy
Also builds worker templates from metaworker templates
"""

from ..HGui import Attribute, Hook
from ..params import get_paramtypelist, parse_paramtypelist
from .WorkerInstanceManager import WorkerInstance
import copy

class WorkerMapping(object):
  def __init__(self):
    self._inmap = {} #Describes the mapping from antennas to attribute inhooks
    self._pmap = {}  #Describes the mapping from parameters to attribute values
    self._outmap = {}#Describes the mapping from outputs to attribute outhooks
    self._inmapr = {} #Describes the mapping from attribute inhooks to antennas
    self._pmapr = {}  #Describes the mapping from attribute values to parameters
    self._outmapr = {}#Describes the mapping from attribute outhooks to outputs

  def set_inmap(self, m):
    self._inmap = m
    r = {}
    for k,v in m.items():
      if v != None: r[v] = k
    self._inmapr = r      
  def get_inmap(self):
    return self._inmap 
  inmap = property(get_inmap, set_inmap)    

  def set_pmap(self, m):
    self._pmap = m
    r = {}
    for k,v in m.items():
      if v != None: r[v] = k
    self._pmapr = r      
  def get_pmap(self):
    return self._pmap 
  pmap = property(get_pmap, set_pmap)    

  def set_outmap(self, m):
    self._outmap = m
    r = {}
    for k,v in m.items():
      if v != None: r[v] = k
    self._outmapr = r      
  def get_outmap(self):
    return self._outmap 
  outmap = property(get_outmap, set_outmap)    

def build_worker_plain(beename, antennas, outputs, ev, paramnames, paramtypelist):
  mapping = WorkerMapping()  
  attribs = []
  attrs = [(pname, "antenna", pval) for pname,pval in antennas.items()] + \
          [(pname, "output", pval) for pname,pval in outputs.items()]
  attrs.sort(key = lambda i:i[0])

  inmap, outmap, pmap = mapping.inmap, mapping.outmap, mapping.pmap
  for pname, io, pval in attrs:
    mode, type_ = pval
    h = Hook(mode, type_)
    if io == "antenna":
      inh, outh = h, None
      inmap[pname] = pname
    else:
      inh, outh = None, h
      outmap[pname] = pname
    a = Attribute (
      pname,
      inh, outh,
    ) 
    attribs.append(a)
  for pname in "evin", "evout", "everr", "evexc":
    if pname not in ev: continue
    if pname == "evin":
      inmap[pname] = None
    else:
      outmap[pname] = None
  for paramname in paramnames:
    pmap[paramname] = None
  mapping.inmap, mapping.outmap, mapping.pmap = inmap, outmap, pmap
  return attribs, mapping

def build_worker_default(beename, antennas, outputs, ev, paramnames, paramtypelist):
  mapping = WorkerMapping()  
  attribs = []
  attrs = [(pname, "antenna", pval) for pname,pval in antennas.items()] + \
          [(pname, "output", pval) for pname,pval in outputs.items()]
  attrs.sort(key = lambda i:i[0])

  inmap, outmap, pmap = mapping.inmap, mapping.outmap, mapping.pmap
  
  has_value = False
  valuetype = None
  for vv in "value", "v":
    if vv in paramnames:
      has_value = True
      d = dict(zip(paramnames,paramtypelist))
      valuetype = d[vv][0][0]
      pmap[vv] = "value"
    
  for pname, io, pval in attrs:
    if io == "antenna" and pname == "inp":
      if has_value == True and pval[1] != valuetype:
        has_value = False
      break
    if io == "output" and pname == "outp":
      if has_value == True and pval[1] != valuetype:
        has_value = False
      break

  if has_value:
    a_value = Attribute("value",type=valuetype)
    attribs.append(a_value)
  for pname, io, pval in attrs:
    mode, type_ = pval
    h = Hook(mode, type_)
    if io == "antenna":
      if has_value and pname == "inp":
        a_value.inhook = h
        inmap[pname] = "value"
        continue
      inh, outh = h, None
      inmap[pname] = pname
    else:
      if has_value and pname == "outp":
        a_value.outhook = h
        outmap[pname] = "value"
        continue
      inh, outh = None, h
      outmap[pname] = pname
    a = Attribute (
      pname,
      inh, outh,
    )     
    attribs.append(a)
  for pname in "evin", "evout", "everr", "evexc":
    if pname not in ev: continue
    if pname == "evin":
      inmap[pname] = None
    else:
      outmap[pname] = None
  for paramname in paramnames:
    if paramname not in pmap:
      pmap[paramname] = None
  mapping.inmap, mapping.outmap, mapping.pmap = inmap, outmap, pmap
  return attribs, mapping


def attribs_append_evio(attribs, mapping, ev):
  inmap, outmap, pmap = mapping.inmap, mapping.outmap, mapping.pmap

  if "evin" in ev:
    h = Hook("push", "event")
    a = Attribute (
      "evin",
      h, None
    )
    inmap["evin"] = "evin"
    attribs.append(a)

  if "evout" in ev:
    h = Hook("push", "event")
    a = Attribute (
      "evout",
      None, h
    )
    outmap["evout"] = "evout"
    attribs.append(a)

  if "everr" in ev:
    h = Hook("push", "event")
    a = Attribute (
      "everr",
      None, h
    )
    outmap["everr"] = "everr"
    attribs.append(a)

  if "evexc" in ev:
    h = Hook("push", "exception")
    a = Attribute (
      "evexc",
      None, h
    )
    outmap["evexc"] = "evexc"
    attribs.append(a)
  mapping.inmap, mapping.outmap, mapping.pmap = inmap, outmap, pmap

def build_worker_plain_evio(*args, **kwargs):
  if "ev" in kwargs: ev = kwargs["ev"]
  else: ev = args[3]
  attribs, mapping = build_worker_plain(*args, **kwargs)
  attribs_append_evio(attribs, mapping, ev)
  return attribs, mapping

def build_worker_default_evio(*args, **kwargs):
  if "ev" in kwargs: ev = kwargs["ev"]
  else: ev = args[3]
  attribs, mapping = build_worker_default(*args, **kwargs)
  attribs_append_evio(attribs, mapping, ev)
  return attribs, mapping

class WorkerTemplate(object):
  _type = "worker"
  _builders = {
   "default": build_worker_default,
   "default_evio" : build_worker_default_evio,
   "plain": build_worker_plain,
   "plain_evio" : build_worker_plain_evio,
  } 
  def __init__(self,  
   beename, antennas, outputs, ev, paramnames, paramtypelist
  ):
    self.beename = beename
    self.antennas = antennas
    self.outputs = outputs
    self.ev = ev
    self.paramnames = paramnames
    self.paramtypelist = paramtypelist
    self.forms = {}
    self.primary_instance = None

    for profile in self._builders:
      builder = self._builders[profile]
      attribs, mapping = builder(
       beename, antennas, outputs, ev, paramnames, paramtypelist
      )
      self.forms[profile] = attribs, mapping
    self.primary_instance = WorkerInstance(self._type, self.forms, self.paramnames, self.paramtypelist)
            
  def get_workerinstance(self):
    #TODO: do we need copies or not? Not if we don't modify the attributes in the profiles!!
    return WorkerInstance(self._type, self.forms, self.paramnames, self.paramtypelist)
    """
    forminstances = {}
    for profile in self._builders:
      attribs0 = self.forms[profile]
      attribs = [copy.copy(a) for a in attribs0]
      forminstances[profile] = attribs
    paramtypelist = copy.deepcopy(self.paramtypelist)
    ret = WorkerInstance(self._type, forminstances, self.paramnames, paramtypelist)
    return ret  
    """


from ..segments.SegmentBuilder import \
 SegmentTemplate, SegmentVariableTemplate, \
 SegmentPushBufferTemplate, SegmentPullBufferTemplate
def build_workertemplate(beename, antennas, outputs, ev, paramnames, paramtypelist):
  args = (beename, antennas, outputs, ev, paramnames, paramtypelist)
  if len(ev):
    return WorkerTemplate(*args)
  elif beename == "variable":
    return SegmentVariableTemplate(*args)
  elif beename == "push_buffer":
    return SegmentPushBufferTemplate(*args)
  elif beename == "pull_buffer":
    return SegmentPullBufferTemplate(*args)
  else:
    return SegmentTemplate(*args)
            
class WorkerBuilder(object):
  def __init__(self):
    self._workers = {}
    self._metaworkers = {}
    self._workerclasses = {}
    
  def _build_worker(self, workername, worker):
    gp = worker.guiparams
    beename = gp["__beename__"]
    antennas = gp.get("antennas", {})
    outputs = gp.get("outputs", {})
    ev = gp.get("__ev__", [])
    parameters = gp.get("parameters", {})
    paramnames, paramtypelist = get_paramtypelist(workername, parameters)

    return beename, antennas, outputs, ev, paramnames, paramtypelist
  
  def build_worker(self, id_, worker):
    assert id_ not in self._workers
      
    w = self._build_worker("worker " + id_, worker)
    paramtypelist = w[5]
    if paramtypelist is None: return False #some error in loading...    
    ww = build_workertemplate(*w)
    self._workers[id_] = ww
        
    return True

  def remove_worker(self, id_):
    self._workers.pop(id_)
    
  def build_metaworker(self, id_, metaworker):
    assert id_ not in self._metaworkers

    paramnames, paramtypelist = get_paramtypelist(
     "metaworker " + id_, metaworker.metaguiparams
     )
    if paramtypelist is None: return False #some error in loading...    
    self._metaworkers[id_] = metaworker, paramnames, paramtypelist

  def get_metaworker(self, id_):
    return self._metaworkers[id_]
    
  def remove_metaworker(self, id_):
    self._metaworkers.pop(id_)
  
  def build_worker_from_meta(self, metaworkername, metaparams):
    mw = self._metaworkers[metaworkername]
    metaworker, mparamnames, mparamtypelist = mw
    arglist = []    
    for mparamname in mparamnames:
      v = metaparams.get(mparamname, None)
      arglist.append(v)
    args = parse_paramtypelist(mparamtypelist, arglist)
    mparams = dict(zip(mparamnames, args))
    
    worker = metaworker(**mparams)
    
    w = self._build_worker(metaworkername, worker)

    paramtypelist = w[5]
    assert paramtypelist is not None

    worker = build_workertemplate(*w)

    return worker, mparamnames, mparamtypelist, mparams
     
  def get_workerinstance(self, id_):
    assert id_ in self._workers
    return self._workers[id_].get_workerinstance()
 
  def has_worker(self, id_):
    return id_ in self._workers     
   
  def has_metaworker(self, id_):
    return id_ in self._metaworkers     

