from __future__ import print_function, absolute_import

"""
Manages the finding and building of worker templates
 and the instantiation of workers and worker parameters
"""
import os, functools
from . import WorkerFinder
from ..params import parse_paramtypelist

class WorkerManager(object):
  _workerfinderclass = WorkerFinder
  def __init__(self, workerbuilder, workerinstancemanager, pmanager, pworkercreator):
    self._workerbuilder = workerbuilder
    self._wim = workerinstancemanager
    self._wim.observers_selection.append(self.gui_selects)
    self._wim.observers_remove.append(self.gui_removes)
    self._pmanager = pmanager
    self._pworkercreator = pworkercreator
    self._workerfinder_global = None
    self._workerfinder_local = None 
    self._worker_parameters = {}
    self._worker_metaparameters = {}

  def _remove_workerfinder(self, wf):
    if wf is not None:
      workers = wf.workers
      for workername in sorted(workers.keys()):
        self._workerbuilder.remove_worker(workername)  
      metaworkers = wf.metaworkers
      for metaworkername in sorted(metaworkers.keys()):
        self._workerbuilder.remove_metaworker(metaworkername)  

  def _find_global_workers(self, locationconfs):
    self._remove_workerfinder(self._workerfinder_global)

    wfc = self._workerfinderclass    
    locationmodules, syspath = \
     wfc.find_locationmodules(locationconfs)
    self._workerfinder_global = wfc(locationmodules, syspath)
  
  def _find_local_workers(self, localconfs, localdir):
    self._remove_workerfinder(self._workerfinder_local)
    
    localconfs = [p for p in localconfs if os.path.exists(p)]
    wfc = self._workerfinderclass
    locationmodules, syspath = \
     wfc.find_locationmodules(localconfs,localdir)
    found = self._workerfinder_global._found_workers
    self._workerfinder_local = wfc(locationmodules, syspath, found)

  def find_global_workers(self, currdir):
    locationconfs = [currdir+os.sep+"locations.conf"]
    self._find_global_workers(locationconfs)
  def find_local_workers(self, localdir = None):
    if localdir is None or localdir == "": localdir = os.getcwd()
    localconfs = [localdir+os.sep+"hivegui.conf"]
    self._find_local_workers(localconfs, localdir)

  def build_workers(self, local):   
    if local:
      wf = self._workerfinder_local
    else:
      wf = self._workerfinder_global
    workers = wf.workers
    for workername in sorted(workers.keys()):
      worker = workers[workername]
      self._workerbuilder.build_worker(workername, worker) 
      self._pworkercreator.append(workername)
    metaworkers = wf.metaworkers
    for metaworkername in sorted(metaworkers.keys()):
      metaworker = metaworkers[metaworkername]
      self._workerbuilder.build_metaworker(metaworkername, metaworker)  
      self._pworkercreator.append(metaworkername)

  def build_worker(self, workername):
    found = False
    for wf in (self._workerfinder_local, self._workerfinder_global):
      if wf is None: continue
      if workername in wf.workers:
        worker = wf.workers[workername]
        self._workerbuilder.build_worker(workername, worker) 
        self._pworkercreator.append(workername)
        found = True
        break
      elif workername in wf.metaworkers:
        metaworker = metaworkers[metaworkername]
        self._workerbuilder.build_metaworker(metaworkername, metaworker)  
        self._pworkercreator.append(metaworkername)
        found = True
        break
      
    if not found: raise KeyError(workername)

  def remove_built_worker(self, workername):
    if self._workerbuilder.has_worker(workername):
      self._workerbuilder.remove_worker(workername)
    elif self._workerbuilder.has_metaworker(workername):
      self._workerbuilder.remove_metaworker(workername)  
    else:
      raise KeyError(workername)
    self._pworkercreator.remove(workername)
  
  def instantiate(self, 
   workerid, workertype, x, y, metaparamvalues = None, paramvalues = None
  ):
    assert workerid not in self._worker_parameters, workerid
    
    if self._workerbuilder.has_worker(workertype):
      wi = self._workerbuilder.get_workerinstance(workertype)
      assert metaparamvalues is None \
        or len(metaparamvalues) == 0
    elif self._workerbuilder.has_metaworker(workertype):
      if metaparamvalues is None:
        metaparamvalues = {}
      tt = self._workerbuilder.build_worker_from_meta(
       workertype, metaparamvalues
      )
      self._worker_metaparameters[workerid] = [workertype, tt[0], tt[1], tt[2], tt[3]]
      worker = tt[0]
      wi = worker.primary_instance
    else: raise KeyError  

    self._wim.add_workerinstance(workerid, wi, x, y)
    
    pvalues = {}
    if paramvalues is not None:
      arglist = []    
      for paramname in wi.paramnames:
        v = paramvalues.get(paramname, None)
        arglist.append(v)
      args = parse_paramtypelist(wi.paramtypelist, arglist)
      for paramname, a in zip(wi.paramnames,args):
        if a is not None: pvalues[paramname] = a
    self._worker_parameters[workerid] = [workertype, wi.paramnames, wi.paramtypelist, pvalues]
    self._wim.set_parameters(workerid, pvalues)
  
  def _select(self, workerids):
    self._pmanager.deselect()
    if workerids is None: return
    if len(workerids) == 1:
      workerid = workerids[0]
      pwin_general = self._pmanager.get_pwindow("general")
      pwin_general.load_paramset(workerid)
      if workerid in self._worker_parameters:       
        self._pwindow_select_worker_params(workerid)
      if workerid in self._worker_metaparameters:
        self._pwindow_select_worker_metaparams(workerid)        

  def select(self, workerids):
    self._wim.select(workerids)
    self._select(workerids)

  def gui_selects(self, workerids):
    self._select(workerids)
      
  def _pwindow_select_worker_params(self, workerid):
    params = self._worker_parameters[workerid]
    up = functools.partial(
      self._update_worker_parameters, workerid
    )
    self._pmanager.select_pwidget(workerid, "params",
     params[1], params[2], params[3], up
    )

  def _pwindow_select_worker_metaparams(self, workerid):
    params = self._worker_metaparameters[workerid]
    inst = functools.partial(
     self.instantiate_from_meta, workerid, params[0]
    )
    buttons = [("Create instance", inst)]
    up = functools.partial(
      self._update_worker_metaparameters, workerid
    )
    self._pmanager.select_pwidget(workerid, "metaparams",
     params[2], params[3], params[4], up, buttons
    )
  
  def get_new_empty_workerid(self):
    n = 0
    while 1:
      n += 1
      id_ = "<empty%d>" % n
      if id_ in self._worker_parameters: continue
      if id_ in self._worker_metaparameters: continue
      return id_

  def get_new_workerid(self, workertype):
    n = 0
    template = workertype.split(".")[-1]
    while 1:
      n += 1
      id_ = "%s%d" % (template, n)
      if id_ in self._worker_parameters: continue
      if id_ in self._worker_metaparameters: continue
      return id_
  
  def meta_empty_instantiate(self, workerid, workertype, x, y):
    t = self._workerbuilder.get_metaworker(workertype)
    self._worker_metaparameters[workerid] = [workertype, None, t[1], t[2], {}]
    self._wim.add_empty(workerid, x, y)

  def instantiate_from_meta(self, workerid, workertype, metaparamvalues):
    if metaparamvalues is None:
      print("Cannot instantiate metaworker '%s' (%s): invalid meta-parameters" \
       % (workerid, workertype))
      return
    try:   
      node = self._wim.get_node(workerid)[0]
    except KeyError: 
      print("Cannot instantiate metaworker '%s' (%s): empty no longer exists" \
       % (workerid, workertype))
      return
    x,y = node.position
    
    #first see if this doesn't raise an exception... 
    self._workerbuilder.build_worker_from_meta(
     workertype, metaparamvalues
    )  
      
    #all set, there should be no exceptions now
    if workerid in self._worker_parameters: #re-instantiation, delete the old worker
      #TODO: re-form connections of the old worker, and restore parameters, if we can
      self.remove(workerid)      
    else: #the worker was only a shell, instantiating for the first time
      self._wim.remove_empty(workerid)
      self._worker_metaparameters.pop(workerid)    
      self._pmanager.delete_pwidget(workerid)
      workerid = self.get_new_workerid(workertype)
            
    self.instantiate(workerid, workertype, x, y, metaparamvalues)
    self.select([workerid])
    
  def create(self, 
   workerid, workertype, x, y, metaparamvalues = None, paramvalues = None
  ):
    assert workerid not in self._worker_parameters, workerid

    if metaparamvalues == {}:
      metaparamvalues = None
    if paramvalues == {}:
      paramvalues = None
        
    if metaparamvalues is None and self._workerbuilder.has_metaworker(workertype):
      assert paramvalues is None
      empty_workerid = self.get_new_empty_workerid()
      self.meta_empty_instantiate(empty_workerid, workertype, x, y)
      return empty_workerid
    else:
      self.instantiate(workerid, workertype, x, y, metaparamvalues, paramvalues)
      return workerid
      
  def _remove(self, workerid):
    self._pmanager.delete_pwidget(workerid)
    self._worker_parameters.pop(workerid, None)
    m = self._worker_metaparameters.pop(workerid, None)
    if m is not None:
      worker = m[1]
    
  def remove(self, workerid):
    self._remove(workerid)
    self._wim.remove_workerinstance(workerid)

  def gui_removes(self, workerid):
    self._remove(workerid)
    self._wim.gui_removes_workerinstance(workerid)

  def _rename_worker(self, old_workerid, new_workerid):
    for dic in (self._worker_parameters, self._worker_metaparameters):
      if old_workerid in dic:
        if new_workerid in dic: return False
    for dic in (self._worker_parameters, self._worker_metaparameters):
      if old_workerid in dic:
        params = dic.pop(old_workerid)
        dic[new_workerid] = params    
    return True      

  def rename_worker(self, old_workerid, new_workerid):
    ok = self._rename_worker(old_workerid, new_workerid)
    if not ok:
      raise ValueError("workerid already exists: '%s'" % new_workerid)
    self._pmanager.rename_pwidget(old_workerid, new_workerid)
    self._wim.rename_workerinstance(old_workerid, new_workerid)

  def gui_renames_worker(self, old_workerid, new_workerid):
    """
    Worker has been renamed by the PGui, not the canvas...
    """    
    ok = self._rename_worker(old_workerid, new_workerid)
    if not ok: return False
    self._wim.rename_workerinstance(old_workerid, new_workerid)
    return True
    
  def _update_worker_parameters(self, workerid, parameters):
    if parameters is None: return #TODO: may not be the sane thing to do
    params_old = self._worker_parameters[workerid][3]
    if params_old is None: params_old = {}
    new_parameters = {}
    for param,value in parameters.items():
      if param not in params_old or value != params_old[param]:
        new_parameters[param] = value
    self._wim.set_parameters(workerid, new_parameters)  
    self._worker_parameters[workerid][3] = parameters  
  
  def _update_worker_metaparameters(self, workerid, parameters):
    if parameters is None: return #TODO: may not be the sane thing to do
    self._worker_metaparameters[workerid][4] = parameters  
  
  def workerids(self):
    return sorted(
      list(
       set(self._worker_metaparameters.keys()).union(
        set(self._worker_parameters.keys())
       ) 
      ) 
     )
    
  def get_parameters(self, workerid):
    assert workerid in self._worker_parameters or \
     workerid in self._worker_metaparameters  
    workertype, params, metaparams = None, None, None
    
    if workerid in self._worker_parameters:
      p = self._worker_parameters[workerid]
      workertype = p[0]
      params = p[3]
      if len(params) == 0: params = None      
    if workerid in self._worker_metaparameters:
      p = self._worker_metaparameters[workerid]
      workertype = p[0]    
      metaparams = p[4]
      if len(metaparams) == 0: metaparams = None      
    return workertype, params, metaparams

  def clear_workers(self):   
    for workerid in self.workerids():
      self.remove(workerid)

  def get_wim(self): return self._wim
  
  def unbuild_local_workers(self):
    wf = self._workerfinder_local
    if wf is None: return
    
    workers = wf.workers
    for workername in sorted(workers.keys()):
      self._workerbuilder.remove_worker(workername)  
      self._pworkercreator.remove(workername)
    metaworkers = wf.metaworkers
    for metaworkername in sorted(metaworkers.keys()):
      self._workerbuilder.remove_metaworker(metaworkername)  
      self._pworkercreator.remove(metaworkername)
    
    self._workerfinder_local = None

